-- do not modify this table
local debug_level = {
    DISABLED = 0,
    LEVEL_1  = 1,
    LEVEL_2  = 2
}

-- set this DEBUG to debug_level.LEVEL_1 to enable printing debug_level info
-- set it to debug_level.LEVEL_2 to enable really verbose printing
-- note: this will be overridden by user's preference settings
local DEBUG = debug_level.LEVEL_1

local default_settings =
{
    debug_level  = DEBUG,
    port         = 8214,
    heur_enabled = false,
}

-- for testing purposes, we want to be able to pass in changes to the defaults
-- from the command line; because you can't set lua preferences from the command
-- line using the '-o' switch (the preferences don't exist until this script is
-- loaded, so the command line thinks they're invalid preferences being set)
-- so we pass them in as command arguments insetad, and handle it here:
local args={...} -- get passed-in args
if args and #args > 0 then
    for _, arg in ipairs(args) do
        local name, value = arg:match("(.+)=(.+)")
        if name and value then
            if tonumber(value) then
                value = tonumber(value)
            elseif value == "true" or value == "TRUE" then
                value = true
            elseif value == "false" or value == "FALSE" then
                value = false
            elseif value == "DISABLED" then
                value = debug_level.DISABLED
            elseif value == "LEVEL_1" then
                value = debug_level.LEVEL_1
            elseif value == "LEVEL_2" then
                value = debug_level.LEVEL_2
            else
                error("invalid commandline argument value")
            end
        else
            error("invalid commandline argument syntax")
        end

        default_settings[name] = value
    end
end

local dprint = function() end
local dprint2 = function() end
local function reset_debug_level()
    if default_settings.debug_level > debug_level.DISABLED then
        dprint = function(...)
            print(table.concat({"Lua:", ...}," "))
        end

        if default_settings.debug_level > debug_level.LEVEL_1 then
            dprint2 = dprint
        end
    end
end
-- call it now
reset_debug_level()

dprint2("Wireshark version = ", get_version())
dprint2("Lua version = ", _VERSION)

----------------------------------------
-- Unfortunately, the older Wireshark/Tshark versions have bugs, and part of the point
-- of this script is to test those bugs are now fixed.  So we need to check the version
-- end error out if it's too old.
local major, minor, micro = get_version():match("(%d+)%.(%d+)%.(%d+)")
if major and tonumber(major) <= 1 and ((tonumber(minor) <= 10) or (tonumber(minor) == 11 and tonumber(micro) < 3)) then
        error(  "Sorry, but your Wireshark/Tshark version ("..get_version()..") is too old for this script!\n"..
                "This script needs Wireshark/Tshark version 1.11.3 or higher.\n" )
end

-- more sanity checking
-- verify we have the ProtoExpert class in wireshark, as that's the newest thing this file uses
assert(ProtoExpert.new, "Wireshark does not have the ProtoExpert class, so it's too old - get the latest 1.11.3 or higher")

----------------------------------------


----------------------------------------
-- creates a Proto object, but doesn't register it yet
local aslan = Proto("aslan","ASLAN Communication Protocol")

-- create a protocol field (but not register it yet)
-- the abbreviation should always have "<myproto>." before the specific abbreviation, to avoid collisions

local msgtypes = {
        [0] = "Aslan Hello Message",
        [1] = "Aslan Context Request",
        [2] = "Aslan Authentication Request",
        [3] = "Aslan Association Response",
        [4] = "Aslan Disassociation Response",
		[5] = "Aslan Signal Response",
		[6] = "Aslan Initialisation Response",
		[7] = "Aslan Context Response",
		[8] = "Aslan Authentication Response",
		[9] = "Aslan Handover Request",
		[10] = "Aslan Release Request",
		[11] = "Aslan Flush Request",
		[12] = "Aslan Acknowledgement",
        [13] = "Aslan Connect/Disconnect",
		[14] = "Aslan AFCP Handover"
}

local pf_msg_id      = ProtoField.uint8("aslan.msg_id", "Message ID", base.DEC, msgtypes)
--local pf_sta_mac	 = ProtoField.uint8("aslan.sta_mac", "STA MAC", base.DEC)
--local pf_sta_rssi	 = ProtoField.uint8("aslan.sta_rssi", "STA RSSI", base.DEC)
--local pf_channel	 = ProtoField.uint8("aslan.channel", "Channel", base.DEC)
--local pf_ssid		 = ProtoField.uint8("aslan.ssid", "SSID", base.hex)
----------------------------------------
-- this actually registers the ProtoFields above, into our new Protocol
aslan.fields = { pf_msg_id }

----------------------------------------
-- create some expert info fields (this is new functionality in 1.11.3)
-- Expert info fields are very similar to proto fields: they're tied to our protocol,
-- they're created in a similar way, and registered by setting a 'experts' field to
-- a table of them just as proto fields were put into the 'aslan.fields' above
-- The old way of creating expert info was to just add it to the tree, but that
-- didn't let the expert info be filterable in wireshark, whereas this way does

-- some error expert info's
local ef_too_short = ProtoExpert.new("aslan.too_short.expert", "ASLAN message too short",
                                     expert.group.MALFORMED, expert.severity.ERROR)
local ef_invalid_id = ProtoExpert.new("aslan.invalid_id.expert", "Invalid message ID",
                                     expert.group.PROTOCOL, expert.severity.ERROR)

-- register them
aslan.experts = { ef_too_short, ef_invalid_id }

----------------------------------------
-- we don't just want to display our protocol's fields, we want to access the value of some of them too!
-- There are several ways to do that.  One is to just parse the buffer contents in Lua code to find
-- the values.  But since ProtoFields actually do the parsing for us, and can be retrieved using Field
-- objects, it's kinda cool to do it that way. So let's create some Fields to extract the values.
-- The following creates the Field objects, but they're not 'registered' until after this script is loaded.
local msg_id_field       = Field.new("aslan.msg_id")

-- here's a little helper function to access the response_field value later.
-- Like any Field retrieval, you can't retrieve a field's value until its value has been
-- set, which won't happen until we actually use our ProtoFields in TreeItem:add() calls.
-- So this isResponse() function can't be used until after the pf_flag_response ProtoField
-- has been used inside the dissector.
--
-- A shorter version of this function would be:
-- local function isResponse() return response_field()() end
-- but I though the below is easier to understand.
local function isHelloMessage()
    local msg_id_fieldinfo = msg_id_field()
    return msg_id_fieldinfo()
end

--------------------------------------------------------------------------------
-- preferences handling stuff
--------------------------------------------------------------------------------

-- a "enum" table for our enum pref, as required by Pref.enum()
-- having the "index" number makes ZERO sense, and is completely illogical
-- but it's what the code has expected it to be for a long time. Ugh.
local debug_pref_enum = {
    { 1,  "Disabled", debug_level.DISABLED },
    { 2,  "Level 1",  debug_level.LEVEL_1  },
    { 3,  "Level 2",  debug_level.LEVEL_2  },
}

aslan.prefs.debug = Pref.enum("Debug", default_settings.debug_level,
                            "The debug printing level", debug_pref_enum)

aslan.prefs.port  = Pref.uint("Port number", default_settings.port,
                            "The UDP port number for ASLAN")

aslan.prefs.heur  = Pref.bool("Heuristic enabled", default_settings.heur_enabled,
                            "Whether heuristic dissection is enabled or not")

----------------------------------------
-- a function for handling prefs being changed
function aslan.prefs_changed()
    dprint2("prefs_changed called")

    default_settings.debug_level  = aslan.prefs.debug
    reset_debug_level()

    default_settings.heur_enabled = aslan.prefs.heur

    if default_settings.port ~= aslan.prefs.port then
        -- remove old one, if not 0
        if default_settings.port ~= 0 then
            dprint2("removing ASLAN from port",default_settings.port)
            DissectorTable.get("udp.port"):remove(default_settings.port, aslan)
        end
        -- set our new default
        default_settings.port = aslan.prefs.port
        -- add new one, if not 0
        if default_settings.port ~= 0 then
            dprint2("adding ASLAN to port",default_settings.port)
            DissectorTable.get("udp.port"):add(default_settings.port, aslan)
        end
    end

end

dprint2("ASLAN Prefs registered")


----------------------------------------
---- some constants for later use ----
-- the ASLAN header size
local ASLAN_HDR_LEN = 1

-- the smallest possible ASLAN query field size
-- has to be at least a label length octet, label character, label null terminator, 2-bytes type and 2-bytes class
local MIN_DATA_LEN = 0


----------------------------------------
-- The following creates the callback function for the dissector.
-- It's the same as doing "aslan.dissector = function (tvbuf,pkt,root)"
-- The 'tvbuf' is a Tvb object, 'pktinfo' is a Pinfo object, and 'root' is a TreeItem object.
-- Whenever Wireshark dissects a packet that our Proto is hooked into, it will call
-- this function and pass it these arguments for the packet it's dissecting.
function aslan.dissector(tvbuf,pktinfo,root)
    dprint2("aslan.dissector called")

    -- set the protocol column to show our protocol name
    pktinfo.cols.protocol:set("ASLAN")

    -- We want to check that the packet size is rational during dissection, so let's get the length of the
    -- packet buffer (Tvb).
    -- Because ASLAN has no additional payload data other than itself, and it rides on UDP without padding,
    -- we can use tvb:len() or tvb:reported_len() here; but I prefer tvb:reported_length_remaining() as it's safer.
    local pktlen = tvbuf:reported_length_remaining()

    -- We start by adding our protocol to the dissection display tree.
    -- A call to tree:add() returns the child created, so we can add more "under" it using that return value.
    -- The second argument is how much of the buffer/packet this added tree item covers/represents - in this
    -- case (ASLAN protocol) that's the remainder of the packet.
    local tree = root:add(aslan, tvbuf:range(0,pktlen))

    -- now let's check it's not too short
    if pktlen < ASLAN_HDR_LEN then
        -- since we're going to add this protocol to a specific UDP port, we're going to
        -- assume packets in this port are our protocol, so the packet being too short is an error
        -- the old way: tree:add_expert_info(PI_MALFORMED, PI_ERROR, "packet too short")
        -- the correct way now:
        tree:add_proto_expert_info(ef_too_short)
        dprint("packet length",pktlen,"too short")
        return
    end

    -- Now let's add our msg type under our aslan protocol tree we just created.
    -- The transaction id starts at offset 0, for 1 byte length.
    tree:add(pf_msg_id, tvbuf:range(0,1))
	
	local msg_id = msg_id_field()()
	
	if msg_id > 14 then
		tree:add_proto_expert_info(ef_invalid_id)
		pktinfo.cols.info:set("Invalid Message ID (".. msg_id ..")")
	else
		pktinfo.cols.info:set(msgtypes[msg_id_field()()])
	end
	
	local asc = ""
	local dec = ""
	local hex = ""
	
	if msg_id == 3 then
		for i = 1, 6 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(1,6), "STA MAC:", hex)
        tree:add(tvbuf(7), "STA Context")
		--for i = 7, 1000 do
		--	local c = tvbuf(i,1):uint()
		--	if hex == "" then
		--		hex = string.format("%.2x", c)
		--	else
		--		hex = string.format("%s:%.2x", hex, c)
		--	end
		--end
		--tree:add(tvbuf(7,1000), "STA Context:", hex)
	
	elseif msg_id == 1 or msg_id == 4 or msg_id == 5 or msg_id == 7 or msg_id == 9 or msg_id == 10 then
		
		for i = 1, 6 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(1,6), "STA MAC:", hex)
		hex = ""
		if  msg_id == 7 or msg_id == 9 then
			for i = 7, 12 do
				local c = tvbuf(i,1):uint()
				if hex == "" then
					hex = string.format("%.2x", c)
				else
					hex = string.format("%s:%.2x", hex, c)
				end
			end
			tree:add(tvbuf(7,6), "BSSID:", hex)
			--if msg_id == 9 then
			--	for i = 14, 1000 do
			--		local c = tvbuf(i,1):uint()
			--		if hex == "" then
			--			hex = string.format("%.2x", c)
			--		else
			--			hex = string.format("%s:%.2x", hex, c)
			--		end
			--	end
			--	tree:add(tvbuf(14,1000), "STA Context:", hex)
			--end
			tree:add(tvbuf(13), "STA Context")
		end
		
		if msg_id == 5 then
			hex = string.format("%.2d", tvbuf(7,1):int())
			-- dBm = (tonumber(hex))/2 - 100
			tree:add(tvbuf(7,1), "STA RSSI:", hex, "dBm")
		end
		
	elseif msg_id == 6 then
		dec = string.format("%.2d", tvbuf(1,1):uint())
		tree:add(tvbuf(1,1), "Channel:", dec)
		
		for i = 2, 16 do
			local c = tvbuf(i,1):uint()

    -- only append printable characters (hide garbage with spaces)
			if c >= 0x20 and c <= 0x7E then
				if asc == "" then
					asc = string.format("%c", c)
				else
					asc = string.format("%s%c", asc, c)
				end
			else
				asc = asc .. "    "
			end
		end
		tree:add(tvbuf(2,15), "SSID:", asc)
	
	elseif msg_id == 12 then
		local c = tvbuf(1,1):uint()
		asc="("
		asc=string.format("%s%d", asc, c)
		asc=string.format("%s)",asc)
		if c == 0x01 then
			tree:add(tvbuf(1,1), "Acknowledgement: Negative", asc)
		elseif c == 0x00 then
			tree:add(tvbuf(1,1), "Acknowledgement: Positive", asc)	
		else
			tree:add(tvbuf(1,1), "Acknowledgement: Error", asc)
		end
    
    elseif msg_id == 13 then
		for i = 1, 4 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2d", c)
			else
				hex = string.format("%s.%.2d", hex, c)
			end
		end
		tree:add(tvbuf(1,4), "AFCP IP:", hex)
		hex = ""
		for i = 5, 6 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s%.2x", hex, c)
			end
		end
		dec = tonumber(hex, 16)
		tree:add(tvbuf(5,2), "AFCP Port:", dec)
		hex = ""
		for i = 7, 12 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(7,6), "WTP MAC:", hex)
		hex = ""
		for i = 13, 18 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(13,6), "STA MAC:", hex)
		local c = tvbuf(19,1):uint()
		asc="("
		asc=string.format("%s%d", asc, c)
		asc=string.format("%s)",asc)
		if c == 0x00 then
			tree:add(tvbuf(19,1), "Flag: Connect", asc)	
		else
			tree:add(tvbuf(19,1), "Flag: Disconnect", asc)
		end
        
	elseif msg_id == 14 then
		for i = 1, 4 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2d", c)
			else
				hex = string.format("%s.%.2d", hex, c)
			end
		end
		tree:add(tvbuf(1,4), "AFCP IP:", hex)
		hex = ""
		for i = 5, 6 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s%.2x", hex, c)
			end
		end
		dec = tonumber(hex, 16)
		tree:add(tvbuf(5,2), "AFCP Port:", dec)
		hex = ""
		for i = 7, 12 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(7,6), "Old WTP MAC:", hex)
		hex = ""
		for i = 13, 18 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(13,6), "New WTP MAC:", hex)
		hex = ""
		for i = 19, 24 do
			local c = tvbuf(i,1):uint()
			if hex == "" then
				hex = string.format("%.2x", c)
			else
				hex = string.format("%s:%.2x", hex, c)
			end
		end
		tree:add(tvbuf(19,6), "STA MAC:", hex)
	end
	
	local pos = ASLAN_HDR_LEN
	
	dprint2("aslan.dissector returning",pos)

    -- tell wireshark how much of tvbuff we dissected
    return pos
end

----------------------------------------
-- we want to have our protocol dissection invoked for a specific UDP port,
-- so get the udp dissector table and add our protocol to it
DissectorTable.get("udp.port"):add(default_settings.port, aslan)
